#ifndef __BUFFPOOL_H_
#define __BUFFPOOL_H_ 
#include <iostream>
#include <string>
#include <cstdio>
#include <queue>
#include <mutex>
#include "zeek/Reporter.h"

namespace SPcap
{
#define MAX_PKT_BUFF_POOL 5000
//Buffer Pool for caching buffers used 
//for copying packets
//Assumptions: Allocations are done by main zeek thread.
// Freeing of buffers are handled by the thread handling writes.

template <typename T>
class BuffPool {
	public:
	BuffPool(uint64_t bufferSize);
	~BuffPool();
	std::unique_ptr<T> Get();
	void Put(std::unique_ptr<T> pb);

	private:
		std::mutex poolMtx;
		std::queue<std::unique_ptr<T>	>	*buffPool; //Source of allocation. Will be swapped with free when empty.
		std::queue<std::unique_ptr<T>	>	*releasePool; //Freed buffers goes here
		std::queue<std::unique_ptr<T>	>	buffPool_1;//buffPool/releasePool points here
		std::queue<std::unique_ptr<T>	>	buffPool_2;//buffPool/releasePool points here
		uint64_t buffer_size;
};

template <typename T>
BuffPool<T>::BuffPool(uint64_t bufferSize)
	{
	buffer_size = bufferSize;
	for (int i=0; i < MAX_PKT_BUFF_POOL; i++)
		buffPool_1.push(std::move(std::unique_ptr<T>(new T(buffer_size))));
	buffPool = &buffPool_1;
	releasePool = &buffPool_2;
	}

template <typename T>
BuffPool<T>::~BuffPool()
	{
	}

template <typename T>
std::unique_ptr<T>  BuffPool<T>::Get() 
	{
	if (buffPool->size())
		{
		std::unique_ptr<T> p = std::move(buffPool->front());
		buffPool->pop();
		return p;
		}

	if (releasePool->size()) 
		{
		//Swap free and alloc pools
		std::lock_guard<std::mutex> lck(poolMtx);
		std::swap(buffPool, releasePool);
		}
	else
		{
		return std::unique_ptr<T>(new T(buffer_size));
		}

	std::unique_ptr<T> p = std::move(buffPool->front());
	buffPool->pop();

	return p;
	}

template <typename T>
void BuffPool<T>::Put(std::unique_ptr<T> pb)
	{
	std::lock_guard<std::mutex> lck(poolMtx);
	if (releasePool->size() > (MAX_PKT_BUFF_POOL)) {
		//unique_ptr should get released with return.
		return;
	}
	releasePool->push(std::move(pb));
	return;
	}
}
#endif
