// See the file "COPYING" in the main distribution directory for copyright.
#ifndef __SPCAP_H_
#define __SPCAP_H_
#include <sys/types.h>
#include <string>
#include <cstdio>
#include <queue>
#include <memory>

#include "zeek/iosource/Packet.h"
#include "zeek/iosource/IOSource.h"
#include "zeek/iosource/PktDumper.h"
#include "zeek/UID.h"
#include "ESMWriter.h"
#include "ESMDirectWriter.h"
#include "FileWriter.h"
#include "Noops.h"
#include "zeek/Reporter.h"
extern "C" {
#include <pcap.h>
#include <sys/mman.h>
}

namespace SPcap
{
//How much to buffer while waiting for trigger
//This needs fine tuning during performance testing/tuning
static const int default_buffer_depth = 3000;

//Various states of smart pcap functionality
typedef enum {
	DISABLED,
	BUFFER = 1 ,
	STORE, //Decision is made to store packets
	IGNORE, //can be set from policy rules,trigger scripts
	FLUSH  // Flush out remaining buffer to writer
} SmartPcapState;

//Supported Writers
typedef enum {
	SP_NOOP , //Free packets after buffering
	SP_NOOP_WRITE , //Free off packets after buffering before writing out
	SP_UID_FILE , //Write to local file as uid.pcap
	SP_ESM ,  //Stream to (E)xternal (S)torage (M)anager
	SP_ESM_DIRECT // Stream to ESM (non-threaded)
} RUNNING_MODE;

class PcapInfo;


//Global Configuration for smart pcap
class Config {
public:
		
	Config();
	virtual ~Config() {};
	//Set/Get Global parameters	
	
	//Set Global limit on how much to buffer
	//while waiting for a trigger.
	//This can be overriden by flow based settings
	void SetBufferDepth(uint64_t depth);
	//Set Global limit on how much to capture
	//while waiting for a trigger.
	//This can be overriden by flow based settings	
	void SetCaptureDepth(uint64_t len);
	//Store pcaps as uid.pcap in dir
	void EnableUIDFileStorage(std::string dir);
	void EnableESMStorage(std::string socket_path); //Stream to ESM
	void EnableESMDirectStorage(std::string socket_path); // Stream to ESM (non-threaded)
	void EnableNoop(int level)
		{
		if (level)
			{
			mode = SP_NOOP_WRITE; 
			} 
		else
			{
			mode = SP_NOOP;
			}
		SPWriter.reset(new NoopWriter(buffer_depth));
		}
	void Disable();//Disable Smart Pcap feature.

	//Retrieve current values.
	uint64_t GetBufferDepth() { return buffer_depth;}
	uint64_t GetCaptureDepth() { return cap_len;}
	std::string GetPcapDir() { return pcap_dir;}
	bool Enabled() { return enabled;}
	bool CaptureUnknown() { return capture_unknown;}
	void SetCaptureUnknown(bool capture, uint32_t rule) {capture_unknown = capture; capture_unknown_rule = rule;}
	RUNNING_MODE Mode() { return mode;}
	void SetWriterTimeout(int sec, int u_sec);

	//Writer 
	void SendToWriter(std::unique_ptr<PacketBuff> pb);
	std::unique_ptr<PacketBuff>  AllocatePacket();
	void FreePacket(std::unique_ptr<PacketBuff> pb);	

	//Controlled by Writer	
	void PausePacketFeed()
		{
		run_state = FlowState::PAUSE;
		}

	void ResumePacketFeed()
		{
		run_state = FlowState::RUN;
		}

	void ReportUnknownPacket(zeek::Packet *packet);
	void UnknownPacketExpirationCheck(double time);

private:
	std::unique_ptr<Writer<PacketBuff> > SPWriter; //Writer Thread
	void Init();
	void ProcessSmartpcapConfig();
	void Enable(RUNNING_MODE r_mode);
	FlowState run_state; //Packet flow to ESM
	uint64_t buffer_depth;//How much to buffer
	uint64_t cap_len; //Maximum default capture (req+rsp)
	std::string pcap_dir; //For commandline argument
	RUNNING_MODE mode;
	bool enabled; //Is spcap enabled?. Controlled from bro_init()
	bool capture_unknown;
	PcapInfo *sp_unknown_info;
	uint32_t capture_unknown_rule;
	uint32_t unknown_interval_duration; //secs
};

//Packet Capture information
//Stored in Conn 
class PcapInfo {
public:
	PcapInfo();
	virtual ~PcapInfo();

	void Init(const zeek::ConnTuple* cid, const zeek::Packet* pkt, zeek::UID *id);
	void Init(const zeek::ConnTuple* cid, const zeek::Packet* pkt, std::string tid);
	//Initialize and Start Buffering
	//Invoked from BIF
	void Init(zeek::Connection *conn, uint64_t rule);

	//Should connection buffer packets and wait for triggers?
	//Based on decision from smart pcap rules.
	bool DoBuffer()
		{
		return (enabled && (state != IGNORE));
		}

	//Add packets to current buffer
	//Start storing packets if capture activated
	//by trigger
	void Buffer(const zeek::Packet *pkt, bool flush);
	//Enable Capture of packets on current flow/conn.
	//For Trigger decisions . Part of BIF
	void EnableCapture();
	//Disable Capture of packets on current flow/conn.
	//For Trigger decisions . Part of BIF
	void DisableCapture();

	//Required mainly to handle unknown lever
	//To handle protocols which are not identified by zeek,
	//capture is started when a rule triggers with "unknown" keyword
	//Later, if a protocol is identified (protocol_confirmation) by zeek,
	//Capture is turned to buffering mode.
	void SwitchToBuffering();


	//Set how many bytes to capture for current flow/conn.
	//This will override global settings.
	//BIF enabled.
	//defer_check will defer any limit checks until buffering is over.
	//Currently used to handled depths set on "unknown" triggers
	void SetCaptureDepth(uint64_t depth, bool defer_checks)
		{
                //Depth checks should accept zero for resetting 
                //depths set by wildcard levers(all,unknown,...)
		cap_depth = depth;
		defer_limit_checks = defer_checks;
		}

	bool Enabled()
		{
		return (state != DISABLED);
		}

	bool Ignored()
		{
		return (state == IGNORE);
		}

	//Number of packets to capture
	//BIF enabled.
	void SetCaptureCount(uint64_t count)
		{
		if (!count)
				return;
		/*Reset packet counter*/
		pkt_counter = 0;
		cap_count = count;
		}

	//Adjust how much to buffer while waiting for trigger
	//This will override global settings.
	//BIF enabled.
	void SetBufferingLimit(uint64_t limit);
					
	//Flush out packets to writer
	//Used for cleanup situations.
	//set force to true for immediate flush.
	//Flush is deferred in certain scenarios to
	//handle "unknown" and "all" levers/triggers.
	void FlushPkts(bool force);

	//Set the first packet ptr
	void SetFirstPacket(const zeek::Packet *pkt);

	void SetUnknownIntervalExpiration(double time);
	double GetUnknownIntervalExpiration();
	void UnknownIntervalExpired();

private:
	bool enabled;
	bool defer_limit_checks;
	uint32_t rule_id; //Smart pcap rule index/id.
	uint64_t max_buff; //Maximum data to buffer
	uint64_t cap_depth; //Maximum data to capture and store.
	uint64_t cap_count; //Number of packets to capture.
	uint64_t pkt_counter; //Packets handled
	uint64_t tx_bytes; //Bytes stored/transmitted to ESM
	uint64_t payload_bytes; //Running total of payload bytes captured
	SmartPcapState state; //Enabled, buffering, storing, flushing,...
	std::unique_ptr<PacketBuff> pbuff; //Packets are stored/buffered here.
	uint64_t flow_depth; //How deep are we into the flow.
	zeek::UID *uid;  //Ptr to conn.uid
	std::string tid; // Time ID for unknown packets.
	//Ptr to first packet in a connection.
	//This is required to buffer the first packet when
	//buffering decisions are made from zeek scripts.
	const zeek::Packet *first_pkt;
	bool first_pkt_copy;
	void Close(); //Flush and send Close command.
	void FClose(); //Send Close command.
	void Send();//Transmit buffers to writer thread.
	bool CheckLimits();//Check for various limits enforced. Depth, count,..
	double unknown_interval_expiration;
};

}
#endif
